"""
######################################################################

Copyright (C) 1999-2018, Michele Cappellari
E-mail: cappellari_at_astro.ox.ac.uk

For details on the method see:
    Cappellari M., 2002, MNRAS, 333, 400
    https://ui.adsabs.harvard.edu/abs/2002MNRAS.333..400C

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your
research, I would appreciate an acknowledgement to use of
`the MGE fitting software developed by Cappellari (2002)'.

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

######################################################################

 NAME:
     MGE_FIT_SECTORS_TWIST

 AUTHOR:
       Michele Cappellari, Astrophysics Sub-department, University of Oxford, UK

 PURPOSE:
       Fit a Multi Gaussian Expansion (MGE) model to a set of galaxy surface
       brightness measurements. The measurements are usually taken along
       sectors with a previous call to the routine SECTORS_PHOTOMETRY_TWIST.
       The MGE model is intended to be used as a parametrization for
       the galaxy surface brightness. All measurements within this program
       are in the instrumental units of PIXELS and COUNTS.
       This routine fits MGE models with varying position angle and common center.

 EXPLANATION:
       Further information on MGE_FIT_SECTORS_TWIST is available in
       Cappellari M., 2002, MNRAS, 333, 400

 CALLING SEQUENCE:
       MGE_FIT_SECTORS_TWIST, Radius, Angle, Counts, Eps,
           NGAUSS=ngauss, SIGMAPSF=[sigma1,sigma2,...], NORMPSF=[norm1,norm2,...],
           SCALE=scale, RBOUNDS=[rmin,rmax], QBOUNDS=[qmin,qmax],
           PABOUNDS=[PAmin,PAmax], /PRINT, /FASTNORM, /NEGATIVE, SOL=sol,
           OUTER_SLOPE=outer_slope

 INPUTS:
       Radius = Vector containing the radius of the surface brightness
               measurements, taken from the galaxy center. This is given
               in units of PIXELS (!) of the high resolution image.
       Angle = Vector containing the polar angle of the surface brightness
               measurements, taken from the galaxy major axis.
       Counts = Vector containing the actual surface brightness measurements
               in COUNTS (!) at the polar coordinates specified by the vectors
               Radius and Angle. These three vectors need to have the same
               number of elements.
       Eps = Estimate for the galaxy `average' ellipticity Eps = 1-b/a = 1-q'

 OUTPUTS:
       No output parameters. The results are printed on the screen, plotted
       in a PS file with the /PRINT keyword and passed with the optional
       keyword SOL

 OPTIONAL INPUT KEYWORDS:
       /NEGATIVE - Set this keyword to allow for negative Gaussians in the fit.
               Use this only if everything else didn't work or if there is clear
               evidence that negative Gaussians are actually needed.
       NGAUSS - Number of Gaussians desired in the MGE fit.
               Typical values are in the range 10-20 (default: 12).
       NORMPSF - This is optional if only a scalar is given for SIGMAPSF,
               otherwise it must contain the normalization of each MGE component
               of the PSF, whose sigma is given by SIGMAPSF. The vector needs to
               have the same number of elements of SIGMAPSF and the condition
               TOTAL(normpsf) = 1 must be verified. In other words the MGE PSF
               needs to be normalized. (default: 1).
       OUTER_SLOPE - This scalar forces the surface brightness profile of
               the MGE model to decrease at least as fast as R^(-OUTER_SLOPE)
               at the largest measured radius (Default: OUTER_SLOPE=2).
       PABOUNDS - Two elements vector giving the minimum and maximum
               position angle PA allowed in the MGE fit.
       /PRINT - Set this keyword to print the best-fitting MGE profiles in IDL.PS
       QBOUNDS - Two elements vector giving the minimum and maximum
               axial ratio Q allowed in the MGE fit.
       RBOUNDS - Two elements vector giving the minimum and maximum sigma
               allowed in the MGE fit. This is in PIXELS units.
       SCALE - the pixel scale in arcsec/pixels. This is only needed for the plots.
               (default: 1)
       SIGMAPSF - Scalar giving the sigma of the PSF of the high resolution
               image (see Cappellari [2002, pg. 406] for details), or vector
               with the sigma of an MGE model for the circular PSF.
               This has to be in pixels, as the vector RADIUS above.
               (Default: no convolution)
         SOL - If this keyword has at least 8 elements in input, the Sigma,
               qObs and PA will be used as starting point for the optimization.
               This is useful for testing but is never needed otherwise.
               The format is described in the OUTPUT KEYWORDS below.

 OPTIONAL OUTPUT KEYWORDS:
         SOL - Output keyword containing a 4xNgauss array with the
               best-fitting solution:
               1) sol[0,*] = TotalCounts, of the Gaussians components.
                  The relation TotalCounts = Height*(2*!PI*Sigma^2*qObs)
                  can be used compute the Gaussian central surface
                  brightness (Height)
               2) sol[1,*] = Sigma, is the dispersion of the best-fitting
                  Gaussians in pixels.
               3) sol[2,*] = qObs, is the observed axial ratio (q' in the 
                  notation of Cappellari 2002) of the best-fitting Gaussian 
                  components.
               4) sol[3,*] = PA, is the observed position angle of the
                  best-fitting Gaussian components. It is measured
                  with respect to the same (arbitrary) system of angular
                  coordinates one has adopted for the input vector Angle.

 EXAMPLE:
       The sequence of commands below was used to generate the complete
       MGE model of the Figures 11-12 in Cappellari (2002).
       1) The FITS file is read, the bad regions are masked and
           the image is sky subtracted;
       2) the photometry along sectors is performed with SECTORS_PHOTOMETRY_TWIST;
       3) the resulting measurements are fitted with MGE_FIT_SECTORS_TWIST;
       4) the contour are printed on a PS file with MGE_PRINT_CONTOURS_TWIST.
       The geometric parameters of the galaxy (eps,ang,xc,yc) were
       previously determined using FIND_GALAXY.

           fits_read, 'ngc5831_f702w_mosaic.fits', img, h

           badpixels = where(img le 0) ; mask the unexposed part of the mosaic
           ;
           ; Mask a galaxy and two bright stars
           ;
           s = SIZE(img)
           dist_circle, r, s[1:2], 1408.09, 357.749
           badpixels = [badpixels,WHERE(r LT 200)]
           dist_circle, r, s[1:2], 539.243, 161.264
           badpixels = [badpixels,WHERE(r LT 50)]
           dist_circle, r, s[1:2], 393.028, 713.681
           badpixels = [badpixels,WHERE(r LT 50)]

           skylevel = 13.0 ; In counts
           img = img - skyLevel
           scale = 0.100   ; WFPC2/WF CCD pixel scale arcsec/pixels

           sigmaPSF = 0.4  ; WF pixels: here I use one single Gaussian PSF
           eps = 0.28      ; These values were measured with FIND_GALAXY
           ang = 141.0     ; major axis PA in the inner regions
           xc = 969
           yc = 974

           sectors_photometry_twist, img, ang, xc, yc, radius, counts, angle, $
               BADPIXELS=badpixelsMosaic, MINLEVEL=2.0

           MGE_fit_sectors_twist, radius, angle, counts, eps, $
               NGAUSS=11, SIGMAPSF=sigmaPSF, SOL=sol, /PRINT, $
               SCALE=scale, QBOUNDS=[0.2,1]

           MGE_print_contours_twist, img, ang, xc, yc, sol, $
               FILE='NGC5831_Mosaic.ps', BINNING=5, MAGRANGE=9.5, $
               SIGMAPSF=sigmaPSF, SCALE=scale

 PROCEDURES USED:
       The following procedures are contained in the main MGE_FIT_SECTORS program.
           MGE_FIT_SECTORS_TWIST_PRINT  -- called during the fit to show the fit profiles
           FITFUNC_MGE_SECTORS_TWIST    -- returns the residuals between data and model

       Other IDL routines needed:
           BVLS  -- Michele Cappellari porting of Lawson & Hanson generalized NNLS
                    http://purl.org/cappellari/idl
           MGE_MPFIT -- Craig Markwardt porting of Levenberg-Marquardt MINPACK-1

 MODIFICATION HISTORY:
       V1.0.0 First implementation, Padova, February 1999, Michele Cappellari
       V2.0.0 Major revisions, Leiden, January 2000, MC
       V3.0.0 Significant changes, Padova, July 2000, MC
       V3.1.0 More robust definition of err in FITFUNC_MGE_SECTORS,
           Leiden, 27 April 2001, MC
       V3.2.0 Graphical changes: always show about 7 sectors on the screen,
           and print plots with shared axes. Leiden, 8 July 2001, MC
       V3.3.0 Added MGE PSF convolution, central pixel integration and changed
           program input parameters to make it independent from SECTORS_PHOTOMETRY
           Leiden, 26 July 2001, MC
       V3.4.0 Added varying position angles fit, August 2001, MC
       V3.5.0 Added /FASTNORM keyword, Leiden, 20 September 2001, MC
       V3.6.0 Updated documentation, Leiden, 8 October 2001, MC
       V3.7.0 Modified implementation with /NEGATIVE keyword
           Leiden, 23 October 2001, MC
       V3.8.0 Added explicit stepsize (STEP) of numerical derivative in
           parinfo structure, after suggestion by Craig B. Markwardt.
           Leiden, 23 February 2002, MC
       V3.8.1 Added compilation options, Leiden 20 May 2002, MC
       V3.8.2 Corrected checking of input parameters when SIGMAPSF is a vector.
           Added ERRMSG keyword to MPFIT call. Leiden, 13 October 2002, MC
       V3.8.3 Force the input parameters to the given bounds if they
           fall outside the required range before starting the fit.
           Leiden, 7 March 2003, MC
       V3.8.4: Use N_ELEMENTS instead of KEYWORD_SET to test
           non-logical keywords. Leiden, 9 May 2003, MC
       V3.8.5: Use updated calling sequence for BVLS. Leiden, 20 March 2004, MC
       V3.9.0: Force the surface brightness of the MGE model to decrease at
           least as R^-2 at the largest measured radius. Leiden, 8 May 2004, MC
       V3.9.1: Make sure this routine uses the Nov/2000 version of Craig Markwardt
           MPFIT which was renamed MGE_MPFIT to prevent potential conflicts with
           more recent versions of the same routine. Vicenza, 23 August 2004, MC.
       V3.9.2: Allow forcing the outer slope of the surface brightness profile of
           the MGE model to decrease at least as R^-n at the largest measured
           radius (cfr. version 3.9). Leiden, 23 October 2004, MC
       V3.9.3 Replaced LOGRANGE keyword in example with the new MAGRANGE.
           MC, Leiden, 1 May 2005
       V3.9.4: Changed axes labels in plots. Leiden, 18 October 2005, MC
       V3.9.5: Use more robust la_least_squares (IDL 5.6) instead of SVDC with
           /NEGATIVE keyword. MC, Oxford, 16 May 2008
       V3.9.6: Force Gaussians smaller than the PSF, which have a degenerate
           axial ratio and PA, to have the same axial ratio and PA as the
           mean of the first two well determined Gaussians.
           MC, Oxford, 24 September 2008
       V3.9.7: Changed definition of output PA to be positive in
           anti clockwise direction. Thanks to Remco van den Bosch
           for pointing out the inconsistency. MC, Oxford, 6 June 2009
       V3.9.8: Small change to the treatment of the innermost unresolved
           Gaussians. MC, Oxford, 24 April 2012
       V3.9.9: Explained optional usage of SOL in input.
           MC, Oxford, 10 January 2013
       V3.9.10: Use ERF instead of obsolete ERRORF. MC, Oxford, 24 May 2017
       V4.0.0: Converted mge_fit_sectors into into mge_fit_sectors_twist.
           MC, Oxford, 27 July 2017
       V4.0.1: Changed imports for mgefit package. MC, Oxford, 17 April 2018
       V4.0.2: Dropped legacy Python 2.7 support. MC, Oxford, 12 May 2018
       V4.0.3: Fixed clock DeprecationWarning in Python 3.7.
           Use SciPy 1.1 `maxiter` keyword in `nnls`. MC, Oxford, 1 October 2018
       V4.0.4: With ``negative=True`` use Numpy linalg.lstsq instead of Scipy
           for a better default in the criterion for rank deficiency.
           Fixed ignoring negative weights in output with ``negative=True``.
           MC, Oxford, 1 October 2020


"""

import numpy as np
import matplotlib.pyplot as plt
from scipy import optimize, linalg, special
from time import perf_counter as clock

from mgefit.cap_mpfit import mpfit

#----------------------------------------------------------------------------

def _linspace_open(a, b, num=50):

    dx = (b - a)/num
    x = a + (0.5 + np.arange(num))*dx

    return x

#-------------------------------------------------------------------------------

class mge_fit_sectors_twist:

    def __init__(self, radius, angle, counts, eps,
                 ngauss=None, negative=False, sigmapsf=0., normpsf=1., scale=1.,
                 rbounds=None, qbounds=None, pabounds=None, quiet=False,
                 outer_slope=4, sol=0, plot=False, fignum=1):
        """
        This is the main routine. It is the only one that has to be called from outside.

        """
        # Starting from here the biggest part of this routine is devoted to input
        # checking and loading of shared parameters into a common block
        #
        assert np.all(counts > 0), 'Input counts must be positive'
        assert radius.size == angle.size == counts.size, 'Input vectors must have the same length'
        outer_slope = np.clip(outer_slope, 1, 4)

        # load data vectors into class attributes
        #
        self.radius = radius
        self.counts = counts
        self.angle = angle
        self.scale = scale
        sol = np.asarray(sol)
        self.sigmaPSF = np.atleast_1d(sigmapsf)
        self.normPSF = np.atleast_1d(normpsf)
        self.negative = negative
        self.scale = scale
        self.quiet = quiet

        nPsf = self.sigmaPSF.size
        if nPsf > 1:
            assert self.normPSF.size == nPsf, 'sigmaPSF and normPSF must have the same length'
            assert round(np.sum(normpsf), 2) == 1, 'Error: PSF not normalized'

        self.sectors = np.unique(angle) # Finds the different position angles

        # Open grid in the range [rminLog,rmaxLog]
        # The logarithmic slope of a Gaussian is slope(R) = -(R/sigma)**2
        # so to have a slope -n one needs to be at R = sqrt(n)*sigma.
        # Below we constrain the largest Gaussian to have sigma < rmax/sqrt(n)
        # to force the surface brightness of the MGE model to decrease
        # at least as fast as R**-n at the largest measured radius.
        #
        if rbounds is None:
            lrbounds = np.log10([np.min(radius), np.max(radius)/np.sqrt(outer_slope)])
        else:
            lrbounds = np.log10([rbounds[0], rbounds[1]/np.sqrt(outer_slope)])

        if qbounds is None:
            qbounds = np.array([0.05, 1.]) # no Gaussians flatter than q=0.05
        else:
            qbounds = np.asarray(qbounds)

        if pabounds is None:
            pabounds = np.array([-90., 90.])
        else:
            pabounds = np.asarray(pabounds)

        # If the smallest intrinsic Gaussian has sigma=0.75*sigmaPSF it will produce an
        # observed Gaussian with sigmaObs=SQRT(sigma**2+sigmaPSF**2)=1.25*sigmaPSF.
        # We require the sigma of the Gaussians to be larger than 0.75*sigmaPSF,
        # or the smallest measured radius, whichever is larger.
        #
        if np.sum(sigmapsf) > 0:
            lrbounds[0] = max(lrbounds[0], np.log10(0.75*np.min(sigmapsf)))

        # Here the actual calculation starts. The path is different depending on whether the
        # user has requested the nonlinear method or the linear one by setting the /LINEAR keyword
        #
        t = clock()
        if ngauss is None:
            ngauss = 15
        else:
            assert ngauss < 35, 'Too many Gaussians'
        if sol.size < 8:
            logsigma = _linspace_open(*lrbounds, num=ngauss)  # open grid
            q = np.full_like(logsigma, 1 - eps)
            pa = np.zeros_like(logsigma)
        else:
            ngauss = sol.size // 4
            logsigma = np.log10(sol[1, :])      # Log(sigma)
            q = sol[2, :]                       # qObs
            pa = sol[3, :]                      # position angle

        parinfo = [{'step':0.01, 'limits':lrbounds, 'limited':[1, 1], 'tied':''}
                    for j in range(3*ngauss)]

        logsigma = logsigma.clip(*lrbounds)
        q = q.clip(*qbounds)
        pa = pa.clip(*pabounds)

        for j in range(ngauss, 2*ngauss):
            parinfo[j]['limits'] = qbounds
        for j in range(2*ngauss, 3*ngauss):
            parinfo[j]['limits'] = pabounds
            parinfo[j]['step'] = 1.0 # degrees: step for numerical derivative

        if quiet:
            iterfunc = None
        else:
            iterfunc = self._print

        pars = np.concatenate([logsigma, q, pa])
        mp = mpfit(self._fitfunc, pars, iterfunct=iterfunc,
                   nprint=10, parinfo=parinfo, quiet=quiet)
        sol = mp.params
        bestNorm = mp.fnorm
        nIter = mp.niter # Use iterations of the longer part

        # Print the results for the nonzero Gaussians sorted by increasing sigma
        logSigma, q, pa = sol.reshape(3, -1)
        w = self.weights != 0
        m = w.sum()
        self.weights, self.gauss, logSigma, q, pa = \
            self.weights[w], self.gauss[:, w], logSigma[w], q[w], pa[w]

        j = np.argsort(logSigma)
        self.sol = np.vstack([self.weights[j], 10**logSigma[j], q[j], pa[j]])
        self.absdev = np.mean(np.abs(self.err))

        # Force Gaussians with minor axis smaller than the PSF, which have
        # a degenerate axial ratio, to have the same axial ratio as the mean
        # of the first two well determined Gaussians. Flux is conserved by
        # PSF convolution so no other changes are required
        #
        sigy = self.sol[1, :]*self.sol[2, :]
        w = sigy < np.min(sigmapsf)
        self.sol[2, w] = np.mean(self.sol[2, ~w][:2])

        if not self.quiet:
            print('############################################')
            print(' Computation time: %.2f seconds' % (clock() - t))
            print('  Total Iterations: ', nIter)
            print(' Nonzero Gaussians: ', m)
            print('  Unused Gaussians: ', ngauss - m)
            print(' Sectors used in the fit: ', self.sectors.size)
            print(' Total number of points fitted: ', radius.size)
            print(' Chi2: %.4g ' % bestNorm)
            print(' STDEV: %.4g ' % np.std(self.err))
            print(' MEANABSDEV: %.4g ' % self.absdev)
            print('#####################################################')
            print('  Total_Counts  Sigma_Pixels      qObs         PA')
            print('#####################################################')
            for sol in self.sol.T:
                print(('{:13.6e}{:#13.6g}{:13.6f}{:13.4f}').format(*sol))
            print('+++++++++++++++++++++++++++++++++++++++++++++++++++++')

        if plot:
            self.plot(fignum)

# -------------------------------------------------------------------------------

    def _print(self, myfunct, logsigma, it, fnorm,
               functkw=None, args=None, parinfo=None, dof=None, quiet=0):
        """
        This is called every NPRINT iterations of MPFIT

        """
        print('Iteration: %d  chi2: %.4g  Nonzero: %d/%d'%
              (it, fnorm, np.sum(self.weights != 0), self.weights.size))

# ----------------------------------------------------------------------------

    def plot(self, fignum=1):
        """
        Produces final best-fitting plot

        """

        # Select an x and y plot range that is the same for all plots
        #
        minrad = np.min(self.radius)*self.scale
        maxrad = np.max(self.radius)*self.scale
        mincnt = np.min(self.counts)
        maxcnt = np.max(self.counts)
        xran = minrad * (maxrad/minrad)**np.array([-0.02, +1.02])
        yran = mincnt * (maxcnt/mincnt)**np.array([-0.05, +1.05])

        n = self.sectors.size
        dn = int(round(n/6.))
        nrows = (n-1)//dn + 1 # integer division

        fig, ax = plt.subplots(nrows, 2, sharex=True, sharey='col', num=fignum)
        fig.subplots_adjust(hspace=0.01)

        fig.text(0.04, 0.5, 'counts', va='center', rotation='vertical')
        fig.text(0.96, 0.5, 'error (%)', va='center', rotation='vertical')

        ax[-1, 0].set_xlabel("arcsec")
        ax[-1, 1].set_xlabel("arcsec")

        row = 0
        for j in range(0, n, dn):
            w = np.nonzero(self.angle == self.sectors[j])[0]
            w = w[np.argsort(self.radius[w])]
            r = self.radius[w]*self.scale
            txt = "$%.f^\circ$" % self.sectors[j]

            ax[row, 0].set_xlim(xran)
            ax[row, 0].set_ylim(yran)
            ax[row, 0].loglog(r, self.counts[w], 'C0o')
            ax[row, 0].loglog(r, self.yfit[w], 'C1', linewidth=2)
            ax[row, 0].text(0.98, 0.95, txt, ha='right', va='top', transform=ax[row, 0].transAxes)
            ax[row, 0].loglog(r, self.gauss[w, :]*self.weights[None, :])

            ax[row, 1].semilogx(r, self.err[w]*100, 'C0o')
            ax[row, 1].axhline(linestyle='--', color='C1', linewidth=2)
            ax[row, 1].yaxis.tick_right()
            ax[row, 1].yaxis.set_label_position("right")
            ax[row, 1].set_ylim([-19.5, 20])

            row += 1

#----------------------------------------------------------------------------

    def _fitfunc(self, pars, fjac=None):

        logsigma, q, pa = pars.reshape(3, -1)

        r2 = (self.radius**2)[:, None, None]
        angle = self.angle[:, None, None]
        sigma2 = (10**(2*logsigma))[None, :, None]
        q2 = (q**2)[None, :, None]
        pa = pa[None, :, None]
        sigmaPSF2 = (self.sigmaPSF**2)[None, None, :]
        normPSF = self.normPSF[None, None, :]

        # Analytic convolution with an MGE circular PSF
        # Equations (4,5) in Cappellari (2002)
        sigmaX = np.sqrt(sigma2 + sigmaPSF2)
        sigmaY = np.sqrt(sigma2*q2 + sigmaPSF2)

        # Normalized (volume=1) 2-dim Gaussian in polar coordinates
        ang = np.radians(angle + pa)
        g = np.exp(-0.5*r2*((np.cos(ang)/sigmaX)**2 + (np.sin(ang)/sigmaY)**2))
        self.gauss = np.sum(normPSF/(2*np.pi*sigmaX*sigmaY)*g, 2)

        # Analytic integral of the MGE on the central pixel.
        # Below we assume the central pixel is aligned with the galaxy axes.
        # This is generally not the case, but the error due to this
        # approximation is negligible in realistic situations.
        w = self.radius < 0.5
        g = special.erf(2**-1.5/sigmaX) * special.erf(2**-1.5/sigmaY)
        self.gauss[w, :] = np.sum(normPSF*g, 2)

        A = self.gauss/self.counts[:, None] # gauss*SQRT(weights) = gauss/y
        b = np.ones_like(self.radius)  # y*SQRT(weights) = 1 <== weights = 1/sigma**2 = 1/y**2

        if self.negative:   # Solution by LAPACK linear least-squares
            self.weights = np.linalg.lstsq(A, b, rcond=None)[0]
        else:               # Solution by NNLS
            self.weights = optimize.nnls(A, b, maxiter=30*q.size)[0]

        self.yfit = self.gauss @ self.weights   # Evaluate predictions by matrix multiplications
        self.err = 1 - self.yfit / self.counts     # relative error: yfit, counts are positive quantities
        self.chi2 = np.sum(self.err**2)  # rnorm**2 = TOTAL(err**2) (this value is only used with the /LINEAR keyword)

        return self.err.astype(np.float32)

#----------------------------------------------------------------------------
